q = {
    "test": {
        "huey_class": "huey.MemoryHuey",
        "results": True,
        "store_none": False,
        "immediate": False,
        "utc": True,
        "blocking": True,
        "consumer": {
            "workers": 1,
            "worker_type": "thread",
            "initial_delay": 0.1,
            "backoff": 1.15,
            "max_delay": 10.0,
            "scheduler_interval": 60,
            "periodic": True,
            "check_worker_health": True,
            "health_check_interval": 300,
        },
    },
}



import logging
import sys

from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.module_loading import autodiscover_modules

from huey.consumer_options import ConsumerConfig
from huey.consumer_options import OptionParserHandler

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Queue consumer. Example usage::

    To start the consumer (note you must export the settings module):

    django-admin.py djangohuey queuename
    """

    help = "Run the queue consumer"
    _type_map = {"int": int, "float": float}

    def add_arguments(self, parser):
        option_handler = OptionParserHandler()
        groups = (
            option_handler.get_logging_options(),
            option_handler.get_worker_options(),
            option_handler.get_scheduler_options(),
        )
        for option_list in groups:
            for short, full, kwargs in option_list:
                if short == "-v":
                    full = "--huey-verbose"
                    short = "-V"
                if "type" in kwargs:
                    kwargs["type"] = self._type_map[kwargs["type"]]
                kwargs.setdefault("default", None)
                parser.add_argument(full, short, **kwargs)

        parser.add_argument(
            "-A",
            "--disable-autoload",
            action="store_true",
            dest="disable_autoload",
            help='Do not autoload "tasks.py"',
        )
        parser.add_argument(
            "--queue",
            action="store",
            dest="queue",
            help="Name of the queue consumer to run",
        )

    def handle(self, *args, **options):
        from django_huey import get_queue, get_queue_name

        # Python 3.8+ on MacOS uses an incompatible multiprocess model. In this
        # case we must explicitly configure mp to use fork().
        if sys.version_info >= (3, 8) and sys.platform == "darwin":
            # Apparently this was causing a "context has already been set"
            # error for some user. We'll just pass and hope for the best.
            # They're apple users so presumably nothing important will be lost.
            import multiprocessing

            try:
                multiprocessing.set_start_method("fork")
            except RuntimeError:
                pass

        queue_name = options.get("queue")
        queue = get_queue(queue_name)
        queue_name = get_queue_name(queue_name)

        consumer_options = {}
        consumer_options.update(self.default_queue_settings(queue_name))

        for key, value in options.items():
            if value is not None:
                consumer_options[key] = value

        consumer_options.setdefault(
            "verbose", consumer_options.pop("huey_verbose", None)
        )

        if not options.get("disable_autoload"):
            autodiscover_modules("tasks")

        logger = logging.getLogger("huey")

        config = ConsumerConfig(**consumer_options)
        config.validate()

        # Only configure the "huey" logger if it has no handlers. For example,
        # some users may configure the huey logger via the Django global
        # logging config. This prevents duplicating log messages:
        if not logger.handlers:
            config.setup_logger(logger)

        consumer = queue.create_consumer(**config.values)
        consumer.run()

    def default_queue_settings(self, queue):
        try:
            return settings.DJANGO_HUEY["queues"][queue].get("consumer", {})
        except AttributeError:
            pass
        return {}



from importlib import import_module
from django_huey.exceptions import ConfigurationError
from django_huey import config


def include(path):
    if isinstance(path, str):
        try:
            queue_module = import_module(path)
        except ModuleNotFoundError:
            raise ConfigurationError(
                f"No module named '{path}'. Review included queues modules in DJANGO_HUEY."
            )
    queues = getattr(queue_module, "queues")
    config.include(queues)
    return queues



import unittest

from django_huey import task, get_queue


class MultiQueueTest(unittest.TestCase):
    def setUp(self):

        get_queue("multi-1").flush()
        get_queue("multi-2").flush()

    def test_task_uses_specified_queue(self):
        @task(queue="multi-1")
        def some_fun():
            pass

        some_fun()

        queue1 = get_queue("multi-1")
        queue2 = get_queue("multi-2")
        self.assertEqual(len(queue1), 1)
        self.assertEqual(len(queue2), 0)

        @task(queue="multi-2")
        def some_other_fun():
            pass

        some_other_fun()

        queue1 = get_queue("multi-1")
        queue2 = get_queue("multi-2")
        self.assertEqual(len(queue1), 1)
        self.assertEqual(len(queue2), 1)

    def test_task_uses_default_queue_if_not_specified(self):
        @task()
        def some_default_func():
            pass

        some_default_func()

        queue1 = get_queue("multi-1")
        queue2 = get_queue("multi-2")
        self.assertEqual(len(queue1), 0)
        self.assertEqual(len(queue2), 1)



import unittest
from huey import RedisHuey, MemoryHuey
from django_huey.exceptions import ConfigurationError
from django_huey.utils import include
from django_huey.config import DjangoHueySettingsReader


class DjangoHueyConfigTests(unittest.TestCase):
    def test_djangohuey_config_with_no_settings(self):

        with self.assertRaises(ConfigurationError) as cm:
            DjangoHueySettingsReader(None)

        self.assertEqual("Error: DJANGO_HUEY must be a dictionary", str(cm.exception))

    def test_djangohuey_configure_does_not_raise_error_when_both_settings_are_defined(
        self,
    ):
        DJANGO_HUEY = {
            "queues": {
                "queuename": {
                    "name": "test",
                    "immediate": True,
                }
            }
        }
        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()

    def test_djangohuey_get_queue_when_queue_is_none(self):
        DJANGO_HUEY = {
            "queues": {
                "queuename": {
                    "name": "test",
                    "immediate": True,
                }
            }
        }
        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()

        with self.assertRaises(ConfigurationError) as cm:
            config.get_queue(None)

        self.assertEqual(
            """
Command djangohuey must receive a --queue parameter or define a default queue in DJANGO_HUEY setting.
i.e.:
python manage.py djangohuey --queue first

or in settings file:

DJANGO_HUEY = {
    'default': 'your-default-queue-name',
    'queues': {
        #Your queues here
    }
}
""",
            str(cm.exception),
        )

    def test_djangohuey_get_queue_when_queue_is_none_and_default_queue_is_defined(self):
        DJANGO_HUEY = {
            "default": "queuename",
            "queues": {
                "queuename": {
                    "name": "queue-default",
                    "immediate": True,
                }
            },
        }
        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()

        queue = config.get_queue(None)

        self.assertEqual(queue.name, "queue-default")

    def test_djangohuey_invalid_default_queue(self):
        DJANGO_HUEY = {
            "default": "invalid-queue",
            "queues": {
                "queuename": {
                    "name": "queue-default",
                    "immediate": True,
                }
            },
        }
        with self.assertRaises(ConfigurationError) as cm:
            DjangoHueySettingsReader(DJANGO_HUEY)

        self.assertEqual(
            "Queue defined as default: invalid-queue, is not configured in DJANGO_HUEY.",
            str(cm.exception),
        )

    def test_djangohuey_configure_when_django_huey_setting_is_defined(self, *args):
        DJANGO_HUEY = {
            "queues": {
                "first": {
                    "huey_class": "huey.RedisHuey",  # Huey implementation to use.
                    "name": "testname",  # Use db name for huey.
                },
                "mails": {
                    "huey_class": "huey.MemoryHuey",  # Huey implementation to use.
                    "name": "testnamememory",  # Use db name for huey.
                },
            }
        }

        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()

        self.assertTrue(isinstance(config.get_queue("first"), RedisHuey))
        self.assertEqual(config.get_queue("first").name, "testname")
        self.assertTrue(isinstance(config.get_queue("mails"), MemoryHuey))
        self.assertEqual(config.get_queue("mails").name, "testnamememory")

    def test_djangohuey_configure_when_django_huey_setting_is_an_object_raises_error(
        self, *args
    ):
        DJANGO_HUEY = object()

        with self.assertRaises(ConfigurationError) as cm:
            DjangoHueySettingsReader(DJANGO_HUEY)

        self.assertEqual("Error: DJANGO_HUEY must be a dictionary", str(cm.exception))

    def test_djangohuey_if_name_is_not_defined_queue_name_is_default(self, *args):
        DJANGO_HUEY = {
            "queues": {
                "first": {
                    "huey_class": "huey.RedisHuey",  # Huey implementation to use.
                },
                "mails": {
                    "huey_class": "huey.MemoryHuey",  # Huey implementation to use.
                },
            }
        }

        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()
        self.assertEqual(config.get_queue("first").name, "first")

        self.assertEqual(config.get_queue("mails").name, "mails")

    def test_djangohuey_with_backend_class(self, *args):
        DJANGO_HUEY = {
            "queues": {
                "first": {
                    "backend_class": "huey.RedisHuey",
                }
            }
        }

        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()
        self.assertTrue(isinstance(config.get_queue("first"), RedisHuey))

    def test_djangohuey_invalid_backend_class(self, *args):
        DJANGO_HUEY = {
            "queues": {
                "first": {
                    "backend_class": "huey.RedisHuey2",
                }
            }
        }

        with self.assertRaises(ConfigurationError) as cm:
            config = DjangoHueySettingsReader(DJANGO_HUEY)
            config.configure()

        self.assertEqual(
            "Error: could not import Huey backend: huey.RedisHuey2", str(cm.exception)
        )

    def test_djangohuey_configure_when_there_are_two_queues_with_the_same_name(
        self, *args
    ):
        DJANGO_HUEY = {
            "queues": {
                "first": {
                    "huey_class": "huey.RedisHuey",
                    "name": "first",
                },
                "mails": {
                    "huey_class": "huey.MemoryHuey",
                    "name": "first",
                },
            }
        }

        config = DjangoHueySettingsReader(DJANGO_HUEY)

        with self.assertRaises(ConfigurationError) as cm:
            config.configure()

        self.assertEqual(
            "There are more than one queue with the name 'first'. Check DJANGO_HUEY in your settings file.",
            str(cm.exception),
        )

    def test_djangohuey_get_queue_when_queue_is_included_from_module(self):
        DJANGO_HUEY = {
            "queues": {
                **include("tests.queuesvalid"),
                "mails": {
                    "huey_class": "huey.MemoryHuey",
                    "name": "first",
                },
            },
        }
        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()
        queue = config.get_queue("test")

        self.assertEqual(queue.name, "test")

    def test_djangohuey_include_module_without_queue_variable_raises_attribute_error(
        self,
    ):
        with self.assertRaises(AttributeError):
            {
                "queues": {**include("tests.queuesinvalid")},
            }

    def test_djangohuey_include_non_existing_module(self):
        with self.assertRaises(ConfigurationError) as cm:
            {
                "queues": {**include("tests.nonexisting")},
            }
        self.assertEqual(
            "No module named 'tests.nonexisting'. Review included queues modules in DJANGO_HUEY.",
            str(cm.exception),
        )



DEBUG = True

DATABASES = {
    "default": {"NAME": "testdatabase", "ENGINE": "django.db.backends.sqlite3"}
}


SECRET_KEY = "django_tests_secret_key"
TIME_ZONE = "America/Chicago"
LANGUAGE_CODE = "en-us"
ADMIN_MEDIA_PREFIX = "/static/admin/"
STATICFILES_DIRS = ()

MIDDLEWARE_CLASSES = []

INSTALLED_APPS = ["django_huey"]

DJANGO_HUEY = {
    "default": "multi-2",
    "queues": {
        "multi-1": {
            "huey_class": "huey.MemoryHuey",  # Huey implementation to use.
            "results": True,  # Store return values of tasks.
            "store_none": False,  # If a task returns None, do not save to results.
            "immediate": False,  # If DEBUG=True, run synchronously.
            "utc": True,  # Use UTC for all times internally.
            "blocking": True,  # Perform blocking pop rather than poll Redis.
            "consumer": {
                "workers": 1,
                "worker_type": "thread",
                "initial_delay": 0.1,  # Smallest polling interval, same as -d.
                "backoff": 1.15,  # Exponential backoff using this rate, -b.
                "max_delay": 10.0,  # Max possible polling interval, -m.
                "scheduler_interval": 60,  # Check schedule every second, -s.
                "periodic": True,  # Enable crontab feature.
                "check_worker_health": True,  # Enable worker health checks.
                "health_check_interval": 300,  # Check worker health every second.
            },
        },
        "multi-2": {
            "huey_class": "huey.MemoryHuey",  # Huey implementation to use.
            "results": True,  # Store return values of tasks.
            "store_none": False,  # If a task returns None, do not save to results.
            "immediate": False,  # If DEBUG=True, run synchronously.
            "utc": True,  # Use UTC for all times internally.
            "blocking": True,  # Perform blocking pop rather than poll Redis.
            "consumer": {
                "workers": 1,
                "worker_type": "thread",
                "initial_delay": 0.1,  # Smallest polling interval, same as -d.
                "backoff": 1.15,  # Exponential backoff using this rate, -b.
                "max_delay": 10.0,  # Max possible polling interval, -m.
                "scheduler_interval": 60,  # Check schedule every second, -s.
                "periodic": True,  # Enable crontab feature.
                "check_worker_health": True,  # Enable worker health checks.
                "health_check_interval": 300,  # Check worker health every second.
            },
        },
    },
}
HUEY = {
    "name": "test",
    "immediate": True,
}



from functools import wraps
from django.conf import settings
from django.db import close_old_connections

from django_huey.config import DjangoHueySettingsReader

DJANGO_HUEY = getattr(settings, "DJANGO_HUEY", None)

config = DjangoHueySettingsReader(DJANGO_HUEY)
config.configure()


def get_close_db_for_queue(queue):
    def close_db(fn):
        """Decorator to be used with tasks that may operate on the database."""

        @wraps(fn)
        def inner(*args, **kwargs):
            instance = get_queue(queue)
            if not instance.immediate:
                close_old_connections()
            try:
                return fn(*args, **kwargs)
            finally:
                if not instance.immediate:
                    close_old_connections()

        return inner

    return close_db


def get_queue(queue):
    return config.get_queue(queue)


def get_queue_name(queue):
    return config.get_queue_name(queue)


def task(*args, queue=None, **kwargs):
    return get_queue(queue).task(*args, **kwargs)


def periodic_task(*args, queue=None, **kwargs):
    return get_queue(queue).periodic_task(*args, **kwargs)


def lock_task(*args, queue=None, **kwargs):
    return get_queue(queue).lock_task(*args, **kwargs)


# Task management.


def enqueue(*args, queue=None, **kwargs):
    return get_queue(queue).enqueue(*args, **kwargs)


def restore(*args, queue=None, **kwargs):
    return get_queue(queue).restore(*args, **kwargs)


def restore_all(*args, queue=None, **kwargs):
    return get_queue(queue).restore_all(*args, **kwargs)


def restore_by_id(*args, queue=None, **kwargs):
    return get_queue(queue).restore_by_id(*args, **kwargs)


def revoke(*args, queue=None, **kwargs):
    return get_queue(queue).revoke(*args, **kwargs)


def revoke_all(*args, queue=None, **kwargs):
    return get_queue(queue).revoke_all(*args, **kwargs)


def revoke_by_id(*args, queue=None, **kwargs):
    return get_queue(queue).revoke_by_id(*args, **kwargs)


def is_revoked(*args, queue=None, **kwargs):
    return get_queue(queue).is_revoked(*args, **kwargs)


def result(*args, queue=None, **kwargs):
    return get_queue(queue).result(*args, **kwargs)


def scheduled(*args, queue=None, **kwargs):
    return get_queue(queue).scheduled(*args, **kwargs)


# Hooks.
def on_startup(*args, queue=None, **kwargs):
    return get_queue(queue).on_startup(*args, **kwargs)


def on_shutdown(*args, queue=None, **kwargs):
    return get_queue(queue).on_shutdown(*args, **kwargs)


def pre_execute(*args, queue=None, **kwargs):
    return get_queue(queue).pre_execute(*args, **kwargs)


def post_execute(*args, queue=None, **kwargs):
    return get_queue(queue).post_execute(*args, **kwargs)


def signal(*args, queue=None, **kwargs):
    return get_queue(queue).signal(*args, **kwargs)


def disconnect_signal(*args, queue=None, **kwargs):
    return get_queue(queue).disconnect_signal(*args, **kwargs)


def db_task(*args, **kwargs):
    queue = kwargs.get("queue")

    def decorator(fn):
        ret = task(*args, **kwargs)(get_close_db_for_queue(queue)(fn))
        ret.call_local = fn
        return ret

    return decorator


def db_periodic_task(*args, **kwargs):
    queue = kwargs.get("queue")

    def decorator(fn):
        ret = periodic_task(*args, **kwargs)(get_close_db_for_queue(queue)(fn))
        ret.call_local = fn
        return ret

    return decorator



import unittest

from unittest import mock
from django_huey import (
    task,
    periodic_task,
    lock_task,
    enqueue,
    restore,
    restore_all,
    restore_by_id,
    revoke,
    revoke_all,
    revoke_by_id,
    is_revoked,
    result,
    scheduled,
    on_startup,
    on_shutdown,
    pre_execute,
    post_execute,
    signal,
    disconnect_signal,
)

DECORATORS = [
    task,
    periodic_task,
    lock_task,
    enqueue,
    restore,
    restore_all,
    restore_by_id,
    revoke,
    revoke_all,
    revoke_by_id,
    is_revoked,
    result,
    scheduled,
    on_startup,
    on_shutdown,
    pre_execute,
    post_execute,
    signal,
    disconnect_signal,
]


class DecoratorsTest(unittest.TestCase):
    def make_test_decorator(self, decorator):
        with mock.patch("django_huey.get_queue") as obj:

            @decorator(queue="test_queue")
            def test_func():
                pass

            obj.assert_called_once_with("test_queue")

    def test_decorators(self):
        for decorator in DECORATORS:
            self.make_test_decorator(decorator)
